/* GStreamer Element
 * Copyright (C) 2016 Mark Nauwelaerts <mnauw@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1307, USA.
 */

/**
 * SECTION:element-alsavolume
 *
 * <refsect2>
 * <para>
 * A stream volume interface element to control an ALSA mixer volume element.
 * It is meant to be used in combination with alsasink to form a bin
 * (alsavolume ! alsavolume) set on playbin's audio-sink property.
 * </refsect2>
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gst/gst.h>
#include <gst/base/gstbasetransform.h>
#include <gst/audio/audio.h>

#include <gst/entrans.h>
#include <string.h>

#include <alsa/asoundlib.h>
#include <errno.h>
#include <poll.h>
#include <sys/time.h>
#include <sys/socket.h>

#define GST_TYPE_ALSA_VOLUME \
  (gst_alsa_volume_get_type())
#define GST_ALSA_VOLUME(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_ALSA_VOLUME,GstAlsaVolume))
#define GST_ALSA_VOLUME_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_ALSA_VOLUME,GstAlsaVolumeClass))
#define GST_IS_ALSA_VOLUME(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_ALSA_VOLUME))
#define GST_IS_ALSA_VOLUME_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_ALSA_VOLUME))


typedef struct _GstAlsaVolume GstAlsaVolume;
typedef struct _GstAlsaVolumeClass GstAlsaVolumeClass;

struct _GstAlsaVolume
{
  GstBaseTransform basetransform;

  snd_mixer_t *handle;
  snd_mixer_elem_t *elem;
  gint control_sock[2];
  GThread *thread;

  /* properties */
  gchar *device;
  gchar *control;
  gboolean mute;
  gboolean mute_set:1;
  gboolean volume_set:1;
  gdouble volume;
};

struct _GstAlsaVolumeClass
{
  GstBaseTransformClass parent_class;
};

GST_DEBUG_CATEGORY_STATIC (alsa_volume_debug);
#define GST_CAT_DEFAULT alsa_volume_debug

/* GstAlsaVolume signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_DEVICE,
  PROP_CONTROL,
  PROP_MUTE,
  PROP_VOLUME,
  /* FILL ME */
};

/* avoid 'default' as that might go to pulseaudio */
#define DEFAULT_DEVICE          "hw:0"
#define DEFAULT_CONTROL         "PCM"
#define DEFAULT_MUTE            FALSE
#define DEFAULT_VOLUME          1.0
/* as is the case elsewhere, we clip internally */
#define VOLUME_MAX_DOUBLE       10.0

static GstStaticPadTemplate gst_alsa_volume_src_template =
GST_STATIC_PAD_TEMPLATE (GST_BASE_TRANSFORM_SRC_NAME,
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS_ANY);

static GstStaticPadTemplate gst_alsa_volume_sink_template =
GST_STATIC_PAD_TEMPLATE (GST_BASE_TRANSFORM_SINK_NAME,
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS_ANY);

static gboolean gst_alsa_volume_start (GstBaseTransform * trans);
static gboolean gst_alsa_volume_stop (GstBaseTransform * trans);
static gboolean gst_alsa_volume_accept_caps (GstBaseTransform * base,
    GstPadDirection direction, GstCaps * caps);

static void gst_alsa_volume_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_alsa_volume_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static void gst_alsa_volume_finalize (GObject * object);

#define parent_class gst_alsa_volume_parent_class
typedef GstAlsaVolumeClass GstEntransAlsaVolumeClass;
typedef GstAlsaVolume GstEntransAlsaVolume;
G_DEFINE_TYPE_WITH_CODE (GstEntransAlsaVolume, gst_alsa_volume,
    GST_TYPE_BASE_TRANSFORM, G_IMPLEMENT_INTERFACE (GST_TYPE_STREAM_VOLUME,
        NULL)
    );

static void
gst_alsa_volume_class_init (GstAlsaVolumeClass * g_class)
{
  GObjectClass *gobject_class;
  GstElementClass *element_class;
  GstBaseTransformClass *trans_class;

  gobject_class = G_OBJECT_CLASS (g_class);
  element_class = GST_ELEMENT_CLASS (g_class);
  trans_class = GST_BASE_TRANSFORM_CLASS (g_class);

  GST_ENTRANS_DEBUG_CATEGORY_INIT (alsa_volume_debug, "alsavolume", 0,
      "alsavolume");

  gobject_class->set_property = gst_alsa_volume_set_property;
  gobject_class->get_property = gst_alsa_volume_get_property;
  gobject_class->finalize = gst_alsa_volume_finalize;

  g_object_class_install_property (gobject_class, PROP_DEVICE,
      g_param_spec_string ("device", "Device",
          "ALSA device, as defined in an asound configuration file",
          DEFAULT_DEVICE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_CONTROL,
      g_param_spec_string ("control", "Control",
          "ALSA mixer control",
          DEFAULT_CONTROL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_override_property (gobject_class, PROP_MUTE, "mute");

  g_object_class_override_property (gobject_class, PROP_VOLUME, "volume");

  gst_element_class_set_static_metadata (element_class, "ALSA Volume",
      "Audio", "Adjust ALSA Volume",
      "Mark Nauwelaerts <mnauw@users.sourceforge.net>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_alsa_volume_sink_template));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_alsa_volume_src_template));

  trans_class->start = GST_DEBUG_FUNCPTR (gst_alsa_volume_start);
  trans_class->stop = GST_DEBUG_FUNCPTR (gst_alsa_volume_stop);
  trans_class->accept_caps = GST_DEBUG_FUNCPTR (gst_alsa_volume_accept_caps);
}

static void
gst_alsa_volume_init (GstAlsaVolume * avolume)
{
  GstBaseTransform *trans = GST_BASE_TRANSFORM (avolume);

  gst_base_transform_set_passthrough (trans, TRUE);
  gst_base_transform_set_gap_aware (trans, TRUE);

  avolume->device = g_strdup (DEFAULT_DEVICE);
  avolume->control = g_strdup (DEFAULT_CONTROL);
  avolume->mute = DEFAULT_MUTE;
  avolume->volume = DEFAULT_VOLUME;
}

static void
gst_alsa_volume_finalize (GObject * object)
{
  GstAlsaVolume *avolume = GST_ALSA_VOLUME (object);

  g_free (avolume->device);
  g_free (avolume->control);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
gst_alsa_volume_close (GstAlsaVolume * self)
{
  if (self->handle) {
    GST_DEBUG_OBJECT (self, "closing mixer handle");
    snd_mixer_close (self->handle);
    self->handle = NULL;
  }
}

static gboolean
gst_alsa_volume_accept_caps (GstBaseTransform * base,
    GstPadDirection direction, GstCaps * caps)
{
  gboolean ret;
  GstPad *pad;

  /* Proxy accept-caps */

  if (direction == GST_PAD_SRC)
    pad = GST_BASE_TRANSFORM_SINK_PAD (base);
  else
    pad = GST_BASE_TRANSFORM_SRC_PAD (base);

  ret = gst_pad_peer_query_accept_caps (pad, caps);

  return ret;
}


#define CHECK(call, callname, error) \
G_STMT_START {             \
  if ((err = call) < 0) {  \
    GST_WARNING_OBJECT (self, "Error %d (%s) calling " #call, err, snd_strerror (err)); \
    name = callname; \
    goto error;            \
  }                        \
} G_STMT_END;

static gboolean
gst_alsa_volume_setup (GstAlsaVolume * self)
{
  snd_mixer_t *handle = NULL;
  gint err;
  const gchar *name;
  snd_mixer_elem_t *elem;
  snd_mixer_selem_id_t *sid;
  snd_mixer_selem_id_alloca (&sid);

  gst_alsa_volume_close (self);

  GST_DEBUG_OBJECT (self, "opening mixer %s", self->device);

  CHECK (snd_mixer_open (&handle, 0), "open", error);
  CHECK (snd_mixer_attach (handle, self->device), "attach", error);
  CHECK (snd_mixer_selem_register (handle, NULL, NULL), "register", error);
  CHECK (snd_mixer_load (handle), "load", error);

  GST_DEBUG_OBJECT (self, "finding simple control %s", self->control);

  snd_mixer_selem_id_set_index (sid, 0);
  snd_mixer_selem_id_set_name (sid, self->control);
  elem = snd_mixer_find_selem (handle, sid);
  if (!elem)
    goto not_found;

  self->handle = handle;
  self->elem = elem;

  return TRUE;

  /* ERRORS */
error:
  {
    GST_ELEMENT_ERROR (self, RESOURCE, OPEN_WRITE,
        (("Could not open mixer device.")),
        ("Mixer '%s' %s error: %s", self->device, name, snd_strerror (err)));
    goto fail;
  }
not_found:
  {
    GST_ELEMENT_ERROR (self, RESOURCE, SETTINGS, (NULL),
        ("Mixer '%s'; no simple control '%s'", self->device, self->control));
    goto fail;
  }
fail:
  {
    if (handle)
      snd_mixer_close (handle);
    return FALSE;
  }
}

/* call with LOCK */
static void
gst_alsa_volume_apply (GstAlsaVolume * self)
{
  snd_mixer_elem_t *elem;

  if (!self->handle)
    goto no_handle;

  elem = self->elem;
  if (self->mute_set && snd_mixer_selem_has_playback_switch (elem)) {
    gint err = snd_mixer_selem_set_playback_switch_all (elem, !self->mute);
    GST_DEBUG_OBJECT (self, "setting switch on %d (result %d)",
        !self->mute, err);
    self->mute_set = FALSE;
  }
  if (self->volume_set && snd_mixer_selem_has_playback_volume (elem)) {
    glong min, max, val;
    gint err = snd_mixer_selem_get_playback_volume_range (elem, &min, &max);
    if (err >= 0) {
      GST_DEBUG_OBJECT (self, "volume range %ld - %ld", min, max);
      /* apply rounding */
      val = (max - min) * self->volume + 0.5 + min;
      err = snd_mixer_selem_set_playback_volume_all (elem, val);
      GST_DEBUG_OBJECT (self, "volume set to %ld (result %d)", val, err);
    }
    self->volume_set = FALSE;
  }

  return;

  /* ERRORS */
no_handle:
  {
    GST_DEBUG_OBJECT (self, "no ALSA handle");
    return;
  }
}

/* call with LOCK */
static void
gst_alsa_volume_update (GstAlsaVolume * self, gboolean notify)
{
  snd_mixer_elem_t *elem;
  gint chn;
  gboolean muted = FALSE;
  gdouble volume = 0.0, delta = 0.0;
  gboolean notify_vol = FALSE, notify_mute = FALSE;

  if (!self->handle)
    goto no_handle;

  elem = self->elem;

  /* muted = if all muted */
  /* volume = max (channel volumes) */
  for (chn = 0; chn < SND_MIXER_SCHN_LAST; chn++) {
    if (snd_mixer_selem_has_playback_channel (elem, chn)) {
      if (snd_mixer_selem_has_playback_switch (elem)) {
        gint val;
        gint err = snd_mixer_selem_get_playback_switch (elem, chn, &val);
        GST_DEBUG_OBJECT (self, "get switch value on %d = %d (result %d)",
            chn, val, err);
        if (err >= 0 && !val)
          muted = TRUE;
      }
      if (snd_mixer_selem_has_playback_volume (elem)) {
        glong min, max, val;
        gint err = snd_mixer_selem_get_playback_volume_range (elem, &min, &max);
        delta = 1 / ((gdouble) max - (gdouble) min);
        if (err >= 0) {
          gdouble valp;

          GST_DEBUG_OBJECT (self, "volume range %ld - %ld", min, max);
          err = snd_mixer_selem_get_playback_volume (elem, chn, &val);
          valp = (gdouble) (val - min) * 100 / (gdouble) (max - min);
          GST_DEBUG_OBJECT (self, "get volume on %d = %ld = %f%% (result %d)",
              chn, val, valp, err);
          if (err >= 0)
            volume = MAX (volume, (gdouble) (valp) / 100);
        }
      }
    }
  }

  if (self->mute != muted) {
    self->mute = muted;
    notify_mute = TRUE;
  }

  if ((self->volume - volume) > delta || (volume - self->volume) > delta) {
    self->volume = volume;
    notify_vol = TRUE;
  }

  /* unlock during notify */
  GST_OBJECT_UNLOCK (self);
  if (notify && notify_mute)
    g_object_notify (G_OBJECT (self), "mute");

  if (notify && notify_vol)
    g_object_notify (G_OBJECT (self), "volume");
  GST_OBJECT_LOCK (self);

  return;

  /* ERRORS */
no_handle:
  {
    GST_DEBUG_OBJECT (self, "no ALSA handle");
    return;
  }
}

/* invoked with LOCK held */
static int
gst_alsa_volume_elem_callback (snd_mixer_elem_t * elem, unsigned int mask)
{
  GstAlsaVolume *self =
      (GstAlsaVolume *) snd_mixer_elem_get_callback_private (elem);

  if (mask & SND_CTL_EVENT_MASK_VALUE) {
    GST_DEBUG_OBJECT (self, "detected modified value");
    gst_alsa_volume_update (self, TRUE);
  }

  return 0;
}

static void *
gst_alsa_volume_loop (void *user_data)
{
  GstAlsaVolume *self = (GstAlsaVolume *) user_data;
  snd_mixer_t *handle = self->handle;
  struct pollfd *pollfds = NULL;
  gint nfds = 0, n, err;
  const gchar *errdesc;

  g_return_val_if_fail (self->handle != NULL && self->elem != NULL, NULL);

  GST_OBJECT_LOCK (self);
  /* arrange for callback when processing events */
  snd_mixer_elem_set_callback (self->elem, gst_alsa_volume_elem_callback);
  snd_mixer_elem_set_callback_private (self->elem, self);

  /* poll loop */
  do {
    n = 1 + snd_mixer_poll_descriptors_count (handle);
    if (n != nfds) {
      g_free (pollfds);
      nfds = n;
      pollfds = g_malloc0 (nfds * sizeof *pollfds);
      pollfds[0].fd = self->control_sock[0];
      pollfds[0].events = POLLIN;
    }
    err = snd_mixer_poll_descriptors (handle, &pollfds[1], nfds - 1);
    if (err < 0) {
      errdesc = "snd_mixer_poll_descriptors";
      goto error;
    }
    GST_OBJECT_UNLOCK (self);

    GST_LOG_OBJECT (self, "doing poll");
    n = poll (pollfds, nfds, -1);

    GST_OBJECT_LOCK (self);
    if (n < 0) {
      if (errno != EINTR && errno != EAGAIN)
        goto sys_error;
    }
    if (pollfds[0].revents & POLLIN)
      goto stopped;
    if (n > 0) {
      gushort revents;

      err = snd_mixer_poll_descriptors_revents (handle, &pollfds[1], nfds - 1,
          &revents);
      if (err < 0) {
        errdesc = "snd_mixer_poll_descriptors_revents";
        goto error;
      }
      if (revents & (POLLERR | POLLNVAL)) {
        errdesc = "";
        goto error;
      } else if (revents & POLLIN) {
        GST_LOG_OBJECT (self, "processing events");
        snd_mixer_handle_events (handle);
      }
    }
  } while (1);

exit:
  /* avoid surprising callback at other times, e.g. at close */
  snd_mixer_elem_set_callback (self->elem, NULL);
  GST_OBJECT_UNLOCK (self);

  g_free (pollfds);
  return NULL;

  /* ERRORS */
error:
  {
    GST_ELEMENT_ERROR (self, RESOURCE, READ, (NULL),
        ("Mixer '%s' %s error: %s", self->device, errdesc, snd_strerror (err)));
    goto exit;
  }
sys_error:
  {
    GST_ELEMENT_ERROR (self, RESOURCE, READ, (NULL), GST_ERROR_SYSTEM);
    goto exit;
  }
stopped:
  {
    GST_DEBUG_OBJECT (self, "Poll stopped");
    goto exit;
  }
}

static gboolean
gst_alsa_volume_start (GstBaseTransform * trans)
{
  GstAlsaVolume *avolume = GST_ALSA_VOLUME (trans);

  /* monitoring thread communication setup */
  if (socketpair (PF_UNIX, SOCK_STREAM, 0, avolume->control_sock) < 0)
    goto no_socket_pair;

  /* ALSA setup */
  gst_alsa_volume_setup (avolume);
  GST_OBJECT_LOCK (avolume);
  gst_alsa_volume_apply (avolume);
  gst_alsa_volume_update (avolume, TRUE);
  GST_OBJECT_UNLOCK (avolume);

  /* launch monitor thread */
  avolume->thread = g_thread_new ("alsavolume", gst_alsa_volume_loop, avolume);

  return TRUE;

  /* ERRORS */
no_socket_pair:
  {
    GST_ELEMENT_ERROR (avolume, RESOURCE, OPEN_READ_WRITE, (NULL),
        GST_ERROR_SYSTEM);
    return FALSE;
  }
}

static gboolean
wake_event (gint fd)
{
  ssize_t num_written;
  while ((num_written = write (fd, "W", 1)) != 1) {
    if (num_written == -1 && errno != EAGAIN && errno != EINTR) {
      g_critical ("failed to send wake event to fd %d: %s", fd,
          strerror (errno));
      return FALSE;
    }
  }
  return TRUE;
}

static gboolean
gst_alsa_volume_stop (GstBaseTransform * trans)
{
  GstAlsaVolume *avolume = GST_ALSA_VOLUME (trans);

  /* halt worker thread */
  wake_event (avolume->control_sock[1]);
  if (avolume->thread) {
    g_thread_join (avolume->thread);
    avolume->thread = NULL;
  }

  /* safe to clean up now */
  gst_alsa_volume_close (avolume);

  if (avolume->control_sock[0] >= 0) {
    close (avolume->control_sock[0]);
    avolume->control_sock[0] = -1;
  }
  if (avolume->control_sock[1] >= 0) {
    close (avolume->control_sock[1]);
    avolume->control_sock[1] = -1;
  }

  return TRUE;
}

static void
gst_alsa_volume_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstAlsaVolume *avolume;

  g_return_if_fail (GST_IS_ALSA_VOLUME (object));
  avolume = GST_ALSA_VOLUME (object);

  switch (prop_id) {
    case PROP_DEVICE:
      g_free (avolume->device);
      avolume->device = g_value_dup_string (value);
      /* setting NULL restores the default device */
      if (avolume->device == NULL) {
        avolume->device = g_strdup (DEFAULT_DEVICE);
      }
      break;
    case PROP_CONTROL:
      g_free (avolume->control);
      avolume->control = g_value_dup_string (value);
      /* setting NULL restores the default device */
      if (avolume->control == NULL) {
        avolume->control = g_strdup (DEFAULT_CONTROL);
      }
      break;
    case PROP_MUTE:
      GST_OBJECT_LOCK (avolume);
      avolume->mute = g_value_get_boolean (value);
      avolume->mute_set = TRUE;
      gst_alsa_volume_apply (avolume);
      /* apply may have only partially succeeded */
      gst_alsa_volume_update (avolume, FALSE);
      GST_OBJECT_UNLOCK (avolume);
      break;
    case PROP_VOLUME:
      GST_OBJECT_LOCK (avolume);
      avolume->volume = g_value_get_double (value);
      avolume->volume = MIN (avolume->volume, 1.0);
      avolume->volume_set = TRUE;
      gst_alsa_volume_apply (avolume);
      /* apply may have only partially succeeded */
      gst_alsa_volume_update (avolume, FALSE);
      GST_OBJECT_UNLOCK (avolume);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_alsa_volume_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstAlsaVolume *avolume;

  g_return_if_fail (GST_IS_ALSA_VOLUME (object));
  avolume = GST_ALSA_VOLUME (object);

  (void) avolume;

  switch (prop_id) {
    case PROP_DEVICE:
      g_value_set_string (value, avolume->device);
      break;
    case PROP_CONTROL:
      g_value_set_string (value, avolume->control);
      break;
    case PROP_MUTE:
      g_value_set_boolean (value, avolume->mute);
      break;
    case PROP_VOLUME:
      g_value_set_double (value, avolume->volume);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static gboolean
plugin_init (GstPlugin * plugin)
{
  return gst_element_register (plugin, GST_ENTRANS_PREFIX "alsavolume",
      GST_RANK_NONE, GST_TYPE_ALSA_VOLUME);
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    entransalsa,
    "plugin for ALSA volume control",
    plugin_init, VERSION, "GPL", GST_PACKAGE_NAME, GST_PACKAGE_ORIGIN);
